@@ -15,7 +15,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2005 - 2008 Red Hat, Inc.
+ * Copyright (C) 2005 - 2010 Red Hat, Inc.
  */
 
 #define _XOPEN_SOURCE
@@ -36,46 +36,63 @@
 
 #include <config.h>
 
-#include "nm-dhcp-manager.h"
+#include "nm-dhcp-dhclient.h"
 #include "nm-utils.h"
+#include "nm-logging.h"
 
+G_DEFINE_TYPE (NMDHCPDhclient, nm_dhcp_dhclient, NM_TYPE_DHCP_CLIENT)
 
-#define NM_DHCP_MANAGER_PID_FILENAME	"dhclient"
-#define NM_DHCP_MANAGER_PID_FILE_EXT	"pid"
+#define NM_DHCP_DHCLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_DHCP_DHCLIENT, NMDHCPDhclientPrivate))
 
 #if defined(TARGET_DEBIAN)
-#define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhcp3"
+#define NM_DHCLIENT_LEASE_DIR           LOCALSTATEDIR "/lib/dhcp3"
 #elif defined(TARGET_SUSE) || defined(TARGET_MANDRIVA)
-#define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhcp"
+#define NM_DHCLIENT_LEASE_DIR           LOCALSTATEDIR "/lib/dhcp"
 #else
-#define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhclient"
+#define NM_DHCLIENT_LEASE_DIR           LOCALSTATEDIR "/lib/dhclient"
 #endif
-#define NM_DHCP_MANAGER_LEASE_FILENAME	"dhclient"
-#define NM_DHCP_MANAGER_LEASE_FILE_EXT	"lease"
 
 #define ACTION_SCRIPT_PATH	LIBEXECDIR "/nm-dhcp-client.action"
 
+typedef struct {
+	const char *path;
+	char *conf_file;
+	char *lease_file;
+	char *pid_file;
+} NMDHCPDhclientPrivate;
 
-static char *
-get_pidfile_for_iface (const char * iface)
+const char *
+nm_dhcp_dhclient_get_path (const char *try_first)
 {
-	return g_strdup_printf ("%s/%s-%s.%s",
-	                        NM_DHCP_MANAGER_RUN_DIR,
-	                        NM_DHCP_MANAGER_PID_FILENAME,
-	                        iface,
-	                        NM_DHCP_MANAGER_PID_FILE_EXT);
+	static const char *dhclient_paths[] = {
+		"/sbin/dhclient",
+		"/usr/sbin/dhclient",
+		"/usr/pkg/sbin/dhclient",
+		"/usr/local/sbin/dhclient",
+		NULL
+	};
+	const char **path = dhclient_paths;
+
+	if (strlen (try_first) && g_file_test (try_first, G_FILE_TEST_EXISTS))
+		return try_first;
+
+	while (*path != NULL) {
+		if (g_file_test (*path, G_FILE_TEST_EXISTS))
+			break;
+		path++;
+	}
+
+	return *path;
 }
 
-
 static char *
-get_leasefile_for_iface (const char * iface, const char *uuid)
+get_leasefile_for_iface (const char * iface, const char *uuid, gboolean ipv6)
 {
-	return g_strdup_printf ("%s/%s-%s-%s.%s",
-	                        NM_DHCP_MANAGER_LEASE_DIR,
-	                        NM_DHCP_MANAGER_LEASE_FILENAME,
+	return g_strdup_printf ("%s/dhclient%s-%s-%s.lease",
+	                        NM_DHCLIENT_LEASE_DIR,
+	                        ipv6 ? "6" : "",
 	                        uuid,
-	                        iface,
-	                        NM_DHCP_MANAGER_LEASE_FILE_EXT);
+	                        iface);
 }
 
 static void
@@ -85,7 +102,7 @@ add_lease_option (GHashTable *hash, char *line)
 
 	spc = strchr (line, ' ');
 	if (!spc) {
-		g_warning ("%s: line '%s' did not contain a space", __func__, line);
+		nm_log_warn (LOGD_DHCP, "DHCP lease file line '%s' did not contain a space", line);
 		return;
 	}
 
@@ -93,8 +110,8 @@ add_lease_option (GHashTable *hash, char *line)
 	if (g_str_has_prefix (line, "option ")) {
 		spc = strchr (spc + 1, ' ');
 		if (!spc) {
-			g_warning ("%s: option line '%s' did not contain a second space",
-			           __func__, line);
+			nm_log_warn (LOGD_DHCP, "DHCP lease file option line '%s' did not contain a second space",
+			             line);
 			return;
 		}
 	}
@@ -119,7 +136,7 @@ add_lease_option (GHashTable *hash, char *line)
 }
 
 GSList *
-nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
+nm_dhcp_dhclient_get_lease_config (const char *iface, const char *uuid)
 {
 	GSList *parsed = NULL, *iter, *leases = NULL;
 	char *contents = NULL;
@@ -127,7 +144,7 @@ nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
 	char **line, **split = NULL;
 	GHashTable *hash = NULL;
 
-	leasefile = get_leasefile_for_iface (iface, uuid);
+	leasefile = get_leasefile_for_iface (iface, uuid, FALSE);
 	if (!leasefile)
 		return NULL;
 
@@ -152,9 +169,9 @@ nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
 		} else if (!strcmp (*line, "lease {")) {
 			/* Beginning of a new lease */
 			if (hash) {
-				g_warning ("%s: lease file %s malformed; new lease started "
-				           "without ending previous lease",
-				           __func__, leasefile);
+				nm_log_warn (LOGD_DHCP, "DHCP lease file %s malformed; new lease started "
+				             "without ending previous lease",
+				             leasefile);
 				g_hash_table_destroy (hash);
 			}
 
@@ -166,9 +183,9 @@ nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
 
 	/* Check if the last lease in the file was properly ended */
 	if (hash) {
-		g_warning ("%s: lease file %s malformed; new lease started "
-		           "without ending previous lease",
-		           __func__, leasefile);
+		nm_log_warn (LOGD_DHCP, "DHCP lease file %s malformed; new lease started "
+		             "without ending previous lease",
+		             leasefile);
 		g_hash_table_destroy (hash);
 		hash = NULL;
 	}
@@ -195,8 +212,8 @@ nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
 
 			/* Read lease expiration (in UTC) */
 			if (!strptime (data, "%w %Y/%m/%d %H:%M:%S", &expire)) {
-				g_warning ("%s: couldn't parse expire time '%s'",
-				           __func__, data);
+				nm_log_warn (LOGD_DHCP, "couldn't parse DHCP lease file expire time '%s'",
+				             data);
 				continue;
 			}
 
@@ -238,27 +255,30 @@ nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
 
 		/* IP4 address */
 		if (!inet_pton (AF_INET, data, &tmp)) {
-			g_warning ("%s: couldn't parse IP4 address '%s'", __func__, data);
+			nm_log_warn (LOGD_DHCP, "couldn't parse DHCP lease file IP4 address '%s'", data);
 			goto error;
 		}
 		nm_ip4_address_set_address (addr, tmp.s_addr);
 
 		/* Netmask */
 		data = g_hash_table_lookup (hash, "option subnet-mask");
-		if (!data)
-			data = "255.255.255.0"; /* FIXME: assume class C? */
-		if (!inet_pton (AF_INET, data, &tmp)) {
-			g_warning ("%s: couldn't parse IP4 subnet mask '%s'", __func__, data);
-			goto error;
+		if (data) {
+			if (!inet_pton (AF_INET, data, &tmp)) {
+				nm_log_warn (LOGD_DHCP, "couldn't parse DHCP lease file IP4 subnet mask '%s'", data);
+				goto error;
+			}
+			prefix = nm_utils_ip4_netmask_to_prefix (tmp.s_addr);
+		} else {
+			/* Get default netmask for the IP according to appropriate class. */
+			prefix = nm_utils_ip4_get_default_prefix (nm_ip4_address_get_address (addr));
 		}
-		prefix = nm_utils_ip4_netmask_to_prefix (tmp.s_addr);
 		nm_ip4_address_set_prefix (addr, prefix);
 
 		/* Gateway */
 		data = g_hash_table_lookup (hash, "option routers");
 		if (data) {
 			if (!inet_pton (AF_INET, data, &tmp)) {
-				g_warning ("%s: couldn't parse IP4 gateway '%s'", __func__, data);
+				nm_log_warn (LOGD_DHCP, "couldn't parse DHCP lease file IP4 gateway '%s'", data);
 				goto error;
 			}
 			nm_ip4_address_set_gateway (addr, tmp.s_addr);
@@ -288,28 +308,39 @@ out:
 #define DHCP_HOSTNAME_FORMAT DHCP_HOSTNAME_TAG " \"%s\"; # added by NetworkManager"
 
 static gboolean
-merge_dhclient_config (NMDHCPDevice *device,
+merge_dhclient_config (const char *iface,
+                       const char *conf_file,
                        NMSettingIP4Config *s_ip4,
                        guint8 *anycast_addr,
-                       const char *contents,
-                       const char *orig,
+                       const char *orig_path,
                        GError **error)
 {
 	GString *new_contents;
+	char *orig_contents = NULL;
 	gboolean success = FALSE;
 
-	g_return_val_if_fail (device != NULL, FALSE);
-	g_return_val_if_fail (device->iface != NULL, FALSE);
-	
+	g_return_val_if_fail (iface != NULL, FALSE);
+	g_return_val_if_fail (conf_file != NULL, FALSE);
+
 	new_contents = g_string_new (_("# Created by NetworkManager\n"));
 
+	if (g_file_test (orig_path, G_FILE_TEST_EXISTS)) {
+		GError *read_error = NULL;
+
+		if (!g_file_get_contents (orig_path, &orig_contents, NULL, &read_error)) {
+			nm_log_warn (LOGD_DHCP, "(%s): error reading dhclient configuration %s: %s",
+			             iface, orig_path, read_error->message);
+			g_error_free (read_error);
+		}
+	}
+
 	/* Add existing options, if any, but ignore stuff NM will replace. */
-	if (contents) {
+	if (orig_contents) {
 		char **lines = NULL, **line;
 
-		g_string_append_printf (new_contents, _("# Merged from %s\n\n"), orig);
+		g_string_append_printf (new_contents, _("# Merged from %s\n\n"), orig_path);
 
-		lines = g_strsplit_set (contents, "\n\r", 0);
+		lines = g_strsplit_set (orig_contents, "\n\r", 0);
 		for (line = lines; lines && *line; line++) {
 			gboolean ignore = FALSE;
 
@@ -334,6 +365,7 @@ merge_dhclient_config (NMDHCPDevice *device,
 
 		if (lines)
 			g_strfreev (lines);
+		g_free (orig_contents);
 	} else
 		g_string_append_c (new_contents, '\n');
 
@@ -374,14 +406,13 @@ merge_dhclient_config (NMDHCPDevice *device,
 		                        " initial-interval 1; \n"
 		                        " anycast-mac ethernet %02x:%02x:%02x:%02x:%02x:%02x;\n"
 		                        "}\n",
-		                        device->iface,
+		                        iface,
 		                        anycast_addr[0], anycast_addr[1],
 		                        anycast_addr[2], anycast_addr[3],
 		                        anycast_addr[4], anycast_addr[5]);
 	}
 
-	if (g_file_set_contents (device->conf_file, new_contents->str, -1, error))
-		success = TRUE;
+	success = g_file_set_contents (conf_file, new_contents->str, -1, error);
 
 	g_string_free (new_contents, TRUE);
 	return success;
@@ -393,17 +424,16 @@ merge_dhclient_config (NMDHCPDevice *device,
  * read their single config file and merge that into a custom per-interface
  * config file along with the NM options.
  */
-static gboolean
-create_dhclient_config (NMDHCPDevice *device,
+static char *
+create_dhclient_config (const char *iface,
                         NMSettingIP4Config *s_ip4,
                         guint8 *dhcp_anycast_addr)
 {
-	char *orig = NULL, *contents = NULL;
+	char *orig = NULL, *tmp, *conf_file = NULL;
 	GError *error = NULL;
 	gboolean success = FALSE;
-	char *tmp;
 
-	g_return_val_if_fail (device != NULL, FALSE);
+	g_return_val_if_fail (iface != NULL, FALSE);
 
 #if defined(TARGET_SUSE)
 	orig = g_strdup (SYSCONFDIR "/dhclient.conf");
@@ -412,41 +442,28 @@ create_dhclient_config (NMDHCPDevice *device,
 #elif defined(TARGET_GENTOO)
 	orig = g_strdup (SYSCONFDIR "/dhcp/dhclient.conf");
 #else
-	orig = g_strdup_printf (SYSCONFDIR "/dhclient-%s.conf", device->iface);
+	orig = g_strdup_printf (SYSCONFDIR "/dhclient-%s.conf", iface);
 #endif
 
 	if (!orig) {
-		nm_warning ("%s: not enough memory for dhclient options.", device->iface);
+		nm_log_warn (LOGD_DHCP, "(%s): not enough memory for dhclient options.", iface);
 		return FALSE;
 	}
 
-	tmp = g_strdup_printf ("nm-dhclient-%s.conf", device->iface);
-	device->conf_file = g_build_filename ("/var", "run", tmp, NULL);
+	tmp = g_strdup_printf ("nm-dhclient-%s.conf", iface);
+	conf_file = g_build_filename ("/var", "run", tmp, NULL);
 	g_free (tmp);
 
-	if (!g_file_test (orig, G_FILE_TEST_EXISTS))
-		goto out;
-
-	if (!g_file_get_contents (orig, &contents, NULL, &error)) {
-		nm_warning ("%s: error reading dhclient configuration %s: %s",
-		            device->iface, orig, error->message);
-		g_error_free (error);
-		goto out;
-	}
-
-out:
 	error = NULL;
-	if (merge_dhclient_config (device, s_ip4, dhcp_anycast_addr, contents, orig, &error))
-		success = TRUE;
-	else {
-		nm_warning ("%s: error creating dhclient configuration: %s",
-		            device->iface, error->message);
+	success = merge_dhclient_config (iface, conf_file, s_ip4, dhcp_anycast_addr, orig, &error);
+	if (!success) {
+		nm_log_warn (LOGD_DHCP, "(%s): error creating dhclient configuration: %s",
+		             iface, error->message);
 		g_error_free (error);
 	}
 
-	g_free (contents);
 	g_free (orig);
-	return success;
+	return conf_file;
 }
 
 
@@ -458,80 +475,136 @@ dhclient_child_setup (gpointer user_data G_GNUC_UNUSED)
 	setpgid (pid, pid);
 }
 
-
-GPid
-nm_dhcp_client_start (NMDHCPDevice *device,
-                      const char *uuid,
-                      NMSettingIP4Config *s_ip4,
-                      guint8 *dhcp_anycast_addr)
+static GPid
+dhclient_start (NMDHCPClient *client,
+                const char *ip_opt,
+                const char *mode_opt)
 {
-	GPtrArray *dhclient_argv = NULL;
-	GPid pid = 0;
+	NMDHCPDhclientPrivate *priv = NM_DHCP_DHCLIENT_GET_PRIVATE (client);
+	GPtrArray *argv = NULL;
+	GPid pid = -1;
 	GError *error = NULL;
-	char *pid_contents = NULL;
+	const char *iface, *uuid;
+	char *binary_name, *cmd_str;
+	gboolean ipv6;
+	guint log_domain;
 
-	if (!g_file_test (DHCP_CLIENT_PATH, G_FILE_TEST_EXISTS)) {
-		nm_warning (DHCP_CLIENT_PATH " does not exist.");
-		goto out;
+	g_return_val_if_fail (priv->pid_file == NULL, -1);
+	g_return_val_if_fail (ip_opt != NULL, -1);
+
+	iface = nm_dhcp_client_get_iface (client);
+	uuid = nm_dhcp_client_get_uuid (client);
+	ipv6 = nm_dhcp_client_get_ipv6 (client);
+
+	log_domain = ipv6 ? LOGD_DHCP6 : LOGD_DHCP4;
+
+	priv->pid_file = g_strdup_printf (LOCALSTATEDIR "/run/dhclient%s-%s.pid",
+	                                  ipv6 ? "6" : "",
+	                                  iface);
+	if (!priv->pid_file) {
+		nm_log_warn (log_domain, "(%s): not enough memory for dhcpcd options.", iface);
+		return -1;
 	}
 
-	device->pid_file = get_pidfile_for_iface (device->iface);
-	if (!device->pid_file) {
-		nm_warning ("%s: not enough memory for dhclient options.", device->iface);
-		goto out;
+	if (!g_file_test (priv->path, G_FILE_TEST_EXISTS)) {
+		nm_log_warn (log_domain, "%s does not exist.", priv->path);
+		return -1;
 	}
 
-	device->lease_file = get_leasefile_for_iface (device->iface, uuid);
-	if (!device->lease_file) {
-		nm_warning ("%s: not enough memory for dhclient options.", device->iface);
-		goto out;
+	/* Kill any existing dhclient from the pidfile */
+	binary_name = g_path_get_basename (priv->path);
+	nm_dhcp_client_stop_existing (priv->pid_file, binary_name);
+	g_free (binary_name);
+
+	priv->lease_file = get_leasefile_for_iface (iface, uuid, ipv6);
+	if (!priv->lease_file) {
+		nm_log_warn (log_domain, "(%s): not enough memory for dhclient options.", iface);
+		return -1;
 	}
 
-	if (!create_dhclient_config (device, s_ip4, dhcp_anycast_addr))
-		goto out;
+	argv = g_ptr_array_new ();
+	g_ptr_array_add (argv, (gpointer) priv->path);
 
-	/* Kill any existing dhclient bound to this interface */
-	if (g_file_get_contents (device->pid_file, &pid_contents, NULL, NULL)) {
-		unsigned long int tmp = strtoul (pid_contents, NULL, 10);
+	g_ptr_array_add (argv, (gpointer) "-d");
 
-		if (!((tmp == ULONG_MAX) && (errno == ERANGE)))
-			nm_dhcp_client_stop (device, (pid_t) tmp);
-		remove (device->pid_file);
+	g_ptr_array_add (argv, (gpointer) ip_opt);
+
+	if (mode_opt)
+		g_ptr_array_add (argv, (gpointer) mode_opt);
+
+	g_ptr_array_add (argv, (gpointer) "-sf");	/* Set script file */
+	g_ptr_array_add (argv, (gpointer) ACTION_SCRIPT_PATH );
+
+	g_ptr_array_add (argv, (gpointer) "-pf");	/* Set pid file */
+	g_ptr_array_add (argv, (gpointer) priv->pid_file);
+
+	g_ptr_array_add (argv, (gpointer) "-lf");	/* Set lease file */
+	g_ptr_array_add (argv, (gpointer) priv->lease_file);
+
+	if (priv->conf_file) {
+		g_ptr_array_add (argv, (gpointer) "-cf");	/* Set interface config file */
+		g_ptr_array_add (argv, (gpointer) priv->conf_file);
 	}
 
-	dhclient_argv = g_ptr_array_new ();
-	g_ptr_array_add (dhclient_argv, (gpointer) DHCP_CLIENT_PATH);
+	g_ptr_array_add (argv, (gpointer) iface);
+	g_ptr_array_add (argv, NULL);
 
-	g_ptr_array_add (dhclient_argv, (gpointer) "-d");
+	cmd_str = g_strjoinv (" ", (gchar **) argv->pdata);
+	nm_log_dbg (log_domain, "running: %s", cmd_str);
+	g_free (cmd_str);
 
-	g_ptr_array_add (dhclient_argv, (gpointer) "-sf");	/* Set script file */
-	g_ptr_array_add (dhclient_argv, (gpointer) ACTION_SCRIPT_PATH );
-
-	g_ptr_array_add (dhclient_argv, (gpointer) "-pf");	/* Set pid file */
-	g_ptr_array_add (dhclient_argv, (gpointer) device->pid_file);
-
-	g_ptr_array_add (dhclient_argv, (gpointer) "-lf");	/* Set lease file */
-	g_ptr_array_add (dhclient_argv, (gpointer) device->lease_file);
-
-	g_ptr_array_add (dhclient_argv, (gpointer) "-cf");	/* Set interface config file */
-	g_ptr_array_add (dhclient_argv, (gpointer) device->conf_file);
-
-	g_ptr_array_add (dhclient_argv, (gpointer) device->iface);
-	g_ptr_array_add (dhclient_argv, NULL);
-
-	if (!g_spawn_async (NULL, (char **) dhclient_argv->pdata, NULL, G_SPAWN_DO_NOT_REAP_CHILD,
+	if (!g_spawn_async (NULL, (char **) argv->pdata, NULL, G_SPAWN_DO_NOT_REAP_CHILD,
 	                    &dhclient_child_setup, NULL, &pid, &error)) {
-		nm_warning ("dhclient failed to start.  error: '%s'", error->message);
+		nm_log_warn (log_domain, "dhclient failed to start: '%s'", error->message);
 		g_error_free (error);
-		goto out;
+		pid = -1;
+	} else
+		nm_log_info (log_domain, "dhclient started with pid %d", pid);
+
+	g_ptr_array_free (argv, TRUE);
+	return pid;
+}
+
+static GPid
+real_ip4_start (NMDHCPClient *client,
+                NMSettingIP4Config *s_ip4,
+                guint8 *dhcp_anycast_addr)
+{
+	NMDHCPDhclientPrivate *priv = NM_DHCP_DHCLIENT_GET_PRIVATE (client);
+	const char *iface;
+
+	iface = nm_dhcp_client_get_iface (client);
+
+	priv->conf_file = create_dhclient_config (iface, s_ip4, dhcp_anycast_addr);
+	if (!priv->conf_file) {
+		nm_log_warn (LOGD_DHCP4, "(%s): error creating dhclient configuration file.", iface);
+		return -1;
 	}
 
-	nm_info ("dhclient started with pid %d", pid);
+	return dhclient_start (client, "-4", NULL);
+}
 
-out:
-	g_free (pid_contents);
-	g_ptr_array_free (dhclient_argv, TRUE);
-	return pid;
+static GPid
+real_ip6_start (NMDHCPClient *client,
+                NMSettingIP6Config *s_ip6,
+                guint8 *dhcp_anycast_addr,
+                gboolean info_only)
+{
+	return dhclient_start (client, "-6", info_only ? "-S" : "-N");
+}
+
+static void
+real_stop (NMDHCPClient *client)
+{
+	NMDHCPDhclientPrivate *priv = NM_DHCP_DHCLIENT_GET_PRIVATE (client);
+
+	/* Chain up to parent */
+	NM_DHCP_CLIENT_CLASS (nm_dhcp_dhclient_parent_class)->stop (client);
+
+	if (priv->conf_file)
+		remove (priv->conf_file);
+	if (priv->pid_file)
+		remove (priv->pid_file);
 }
 
 static const char **
@@ -595,10 +668,11 @@ error:
 	return o;
 }
 
-gboolean
-nm_dhcp_client_process_classless_routes (GHashTable *options,
-                                         NMIP4Config *ip4_config,
-                                         guint32 *gwaddr)
+static gboolean
+real_ip4_process_classless_routes (NMDHCPClient *client,
+                                   GHashTable *options,
+                                   NMIP4Config *ip4_config,
+                                   guint32 *gwaddr)
 {
 	const char *str;
 	char **octets, **o;
@@ -624,7 +698,7 @@ nm_dhcp_client_process_classless_routes (GHashTable *options,
 
 	o = octets = g_strsplit (str, " ", 0);
 	if (g_strv_length (octets) < 5) {
-		nm_warning ("Ignoring invalid classless static routes '%s'", str);
+		nm_log_warn (LOGD_DHCP4, "ignoring invalid classless static routes '%s'", str);
 		goto out;
 	}
 
@@ -632,7 +706,7 @@ nm_dhcp_client_process_classless_routes (GHashTable *options,
 		route = NULL;
 		o = (char **) process_rfc3442_route ((const char **) o, &route);
 		if (!route) {
-			nm_warning ("Ignoring invalid classless static routes");
+			nm_log_warn (LOGD_DHCP4, "ignoring invalid classless static routes");
 			break;
 		}
 
@@ -653,8 +727,8 @@ nm_dhcp_client_process_classless_routes (GHashTable *options,
 			inet_ntop (AF_INET, &tmp, addr, sizeof (addr));
 			tmp.s_addr = nm_ip4_route_get_next_hop (route);
 			inet_ntop (AF_INET, &tmp, nh, sizeof (nh));
-			nm_info ("  classless static route %s/%d gw %s",
-			         addr, nm_ip4_route_get_prefix (route), nh);
+			nm_log_info (LOGD_DHCP4, "  classless static route %s/%d gw %s",
+			             addr, nm_ip4_route_get_prefix (route), nh);
 		}
 	}
 
@@ -663,3 +737,42 @@ out:
 	return have_routes;
 }
 
+/***************************************************/
+
+static void
+nm_dhcp_dhclient_init (NMDHCPDhclient *self)
+{
+	NMDHCPDhclientPrivate *priv = NM_DHCP_DHCLIENT_GET_PRIVATE (self);
+
+	priv->path = nm_dhcp_dhclient_get_path (DHCLIENT_PATH);
+}
+
+static void
+dispose (GObject *object)
+{
+	NMDHCPDhclientPrivate *priv = NM_DHCP_DHCLIENT_GET_PRIVATE (object);
+
+	g_free (priv->pid_file);
+	g_free (priv->conf_file);
+	g_free (priv->lease_file);
+
+	G_OBJECT_CLASS (nm_dhcp_dhclient_parent_class)->dispose (object);
+}
+
+static void
+nm_dhcp_dhclient_class_init (NMDHCPDhclientClass *dhclient_class)
+{
+	NMDHCPClientClass *client_class = NM_DHCP_CLIENT_CLASS (dhclient_class);
+	GObjectClass *object_class = G_OBJECT_CLASS (dhclient_class);
+
+	g_type_class_add_private (dhclient_class, sizeof (NMDHCPDhclientPrivate));
+
+	/* virtual methods */
+	object_class->dispose = dispose;
+
+	client_class->ip4_start = real_ip4_start;
+	client_class->ip6_start = real_ip6_start;
+	client_class->stop = real_stop;
+	client_class->ip4_process_classless_routes = real_ip4_process_classless_routes;
+}
+
